//
// This semaphore implementation is fair up-to spurious wakeup.
// This means that if there is no spurious wakeup then it
// behaves fairly.  If spurious wakeup happens, the woken process
// loses its fair position in the notZero condition variable
// queue, and is put at the end of the queue or can acuire the
// the semaphore.
// On the positive side, this solution does not require the use
// of signalAll, and thus it is more efficient.
//

import java.util.*;
import java.util.concurrent.locks.*;

public class FairSemaphore {

    private int sv;
    private final Lock lock = new ReentrantLock(true);
    private final Condition notZero = lock.newCondition();
    private Set<Thread> w = new HashSet<Thread>();

    public FairSemaphore(int sv) {
	this.sv = sv;
    }

    public void P() throws InterruptedException {
	lock.lock();
	try {
	    Thread ct = Thread.currentThread();
	    nap(250);
	    while (sv==0 || (!w.isEmpty() && !w.contains(ct))) {
		w.add(ct);
		System.out.println(Thread.currentThread().getName()+" going to wait");
		notZero.await();
	    }
	    w.remove(ct);
	    sv--;
	}
	finally {
	    lock.unlock();
	}
    }

    private int count = 0;

    public void V() {
	lock.lock();
	try {
	    if (count==0)
		nap(500);
	    count++;
	    sv++;
	    notZero.signal();
	}
	finally {
	    System.out.println("V: "+Thread.currentThread().getName()+" done V, waiters: "+w.size());
	    lock.unlock();
	}
    }

    public static void nap(int ms) {
	try {
	    Thread.sleep(ms);
	}
	catch (InterruptedException e) {
	}
    }

    public static void main(String[] args) {
	final FairSemaphore s = new FairSemaphore(0);

	(new Thread("T1") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T2") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T3") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T4") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T5") {
		public void run() {
		    try {
			nap(1510);
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();

	nap(1500);
	System.out.println("Main going for V");
	s.V();
	System.out.println("Main done V");
    }
 
}
