//
// This implementation is really fair.  It uses internal
// queue of threads to guarantee fair behaviour even in the
// presence of spurious wakeup.  The price is the required
// use of signalAll.
//

import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.locks.*;

public class ReallyFairSemaphore {

    private int sv;
    private final Lock lock = new ReentrantLock(true);
    private final Condition notZero = lock.newCondition();
    private Queue<Thread> w = new ArrayDeque<Thread>();

    private boolean flag = true;

    public ReallyFairSemaphore(int sv) {
	this.sv = sv;
    }

    public void P() throws InterruptedException {
	lock.lock();
	try {
	    Thread ct = Thread.currentThread();
	    nap(250);
            w.add(ct);
	    while (sv==0 || !ct.equals(w.peek())) {
		System.out.println(ct.getName()+" going to wait");
		// Simulate one spurious wakeup for thread T3.
		if (ct.getName().equals("T3") && flag) {
		    System.out.println(ct.getName()+" going for OOPS");
		    notZero.await(1000, TimeUnit.MILLISECONDS);
		    System.out.println(ct.getName()+" OOPS");
		    flag = false;
		}
		else
		    notZero.await();
	    }
	    w.poll();
	    sv--;
	}
	finally {
	    lock.unlock();
	}
    }

    private int count = 0;

    public void V() {
	lock.lock();
	try {
	    if (count==0)
		nap(500);
	    count++;
	    sv++;
	    notZero.signalAll();
	}
	finally {
	    System.out.println("V: "+Thread.currentThread().getName()+" done V, waiters: "+w.size());
	    lock.unlock();
	}
    }

    public static void nap(int ms) {
	try {
	    Thread.sleep(ms);
	}
	catch (InterruptedException e) {
	}
    }

    public static void main(String[] args) {
	final ReallyFairSemaphore s = new ReallyFairSemaphore(0);

	(new Thread("T1") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T2") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T3") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T4") {
		public void run() {
		    try {
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			nap(1000);
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();
	(new Thread("T5") {
		public void run() {
		    try {
			nap(1510);
			System.out.println(getName()+" going for P");
			s.P();
			System.out.println(getName()+" done P");
			s.V();
			System.out.println(getName()+" done V");
		    }
		    catch (InterruptedException e) {}
		}
	    }).start();

	nap(1500);
	System.out.println("Main going for V");
	s.V();
	System.out.println("Main done V");
    }
 
}
